import os
import hiero
import shot_manager_hiero

class SubmitForClientReviewExporter(hiero.exporters.FnTranscodeExporter.TranscodeExporter):
    """Export Processor that does the heavy lifting.
    In this case we just subclass teh default processor and let it do it's thing, then adding some custom code.
    to log info about the output in a spreadseheet.
    """
    def __init__(self, initDict):
        """Initialise exporter.
        initDict is a dicionary that holds a bunch of info about the the task including "item",
        which is the item being processed (e.g. track item, sequence or bin item) nd "preset",
        which holds all teh processor's properties.
        """
        super(SubmitForClientReviewExporter, self).__init__(initDict)
        self.initDict = initDict
        
        with open('/tmp/test.log', 'a') as fh:
            fh.write('\n__init__: {}'.format(self._item))
        
        # store the resolved output path as an attribute for later use
        self.path = self.resolvedExportPath()
        # use the same tokens as the export template to get some additional values (i.e. "{filehead}")
        resolver = self.initDict['resolver']
    
        # get and resolve the "filehead" token
        self.file_name = resolver.resolve(self, '{filehead}')

        # get the root directory from the template
        root = self.initDict['exportRoot']
        # get the shot path form the template (e.g.: {YYYY}_{MM}_{DD}/{shot}_{filehead}.mov) 
        shot_path = self.initDict['shotPath']        
        # join root and shot path and resolve the export tokens
        self.output_path = resolver.resolve(self, os.path.join(root, shot_path))
    
        # get thumbnail path
        self.thumbnail_path = ''
        
    def finishTask(self):
        """Run after each task finishes"""
        with open('/tmp/test.log', 'a') as fh:
            fh.write('\nfinishTask start: {}'.format(self._item))        

        with open('/tmp/test.log', 'a') as fh:
            fh.write('\nfinishTask end: {}'.format(self._item))        
        # now do custom stuff
        if self._preset.properties()['submitThumbnails']:
            # if the custom widget "Submit Thumbnails" is checked save a thumbnail and use it in tehe submission
            self.saveThumbnail()
        self.logToSpreadsheet()
        
        # let the default behaviour clean up
        super(SubmitForClientReviewExporter, self).finishTask()
        
    def startTask(self):
        """Run the default processor, then our some custom code"""
        # call the base implementation
        with open('/tmp/test.log', 'a') as fh:
            fh.write('\nstartTask start: {}'.format(self._item))                
        hiero.exporters.FnTranscodeExporter.TranscodeExporter.startTask(self)
        with open('/tmp/test.log', 'a') as fh:
            fh.write('\startTask end: {}'.format(self._item))        

        # now check if the database handle exists, otherwise throw an error
        self.db_handle = shot_manager_hiero.get_db_handle(silent=True)
        if not self.db_handle:
            self.setError("Error starting transcode\nMissing database connection.\nOpen the Database Settings panel and load a valid xlsx file.")
            hiero.core.log.exception("TranscodeExporter.startTask")


    def logToSpreadsheet(self):
        """Collect data form the item that is currently processed and write it to a spreadsheet"""      
        clip = self._item.source()
        media = clip.mediaSource()
        input_path = media.fileinfos()[0].filename() # use only first fragment for simplicity
        
        start, end = self.outputRange()
        version_data = {'Version Name': self.file_name,
                       'Thumbnail': self.thumbnail_path,
                       'Frame Range': '{}-{}'.format(start, end),
                       'Duration': end - start + 1,
                       'Path To Frames': input_path,
                       'Path To Mov': self.output_path}

        self.db_handle.write_version_data(version_data, sheet_name='playlist')

    def saveThumbnail(self):
        self.thumbnail_path = os.path.splitext(self.output_path)[0] + '_thumb.jpg'
        img = self._item.thumbnail()
        img_small = img.scaledToHeight(180)
        img_small.save(self.thumbnail_path)

class SubmitForClientReviewPreset(hiero.exporters.FnTranscodeExporter.TranscodePreset):
    """The preset class provides access to the exporters properties,
    which the UI class is providing an interface for
    """
    def __init__(self, name, properties):
        super(SubmitForClientReviewPreset, self).__init__(name, properties)
        self._parentType = SubmitForClientReviewExporter

        # extra properties
        self.properties()['submitThumbnails'] = True
        
        # Update properties dictionary with existing data
        self.properties().update(properties)

    def supportedItems(self):
        """Define the types that are supported by ths processor.
        In this case we only want to support track items.
        """
        return hiero.core.TaskPresetBase.kTrackItem

hiero.core.taskRegistry.registerTask(SubmitForClientReviewPreset, SubmitForClientReviewExporter)
