################################################################################################
# A custom menu and actions foruse in Hiero/Nuke Studio
################################################################################################
import os
import hiero
import shot_manager_hiero
import shot_manager_xlsx
from PySide2 import QtWidgets
from PySide2 import QtCore
from shot_manager_xlsx import DB_IO

class DatabaseSettingsPanel(QtWidgets.QWidget):
    """Simple widget to add to Hiero's window manager, so it can be docked like native panels."""

    def __init__(self, parent=None):
        """Initialise the widget with previously saved settings"""
        super(DatabaseSettingsPanel, self).__init__(parent)
        self.setWindowTitle('Database Settings')
        settings_path = os.path.expanduser('~/.nuke/shotManager.ini')
        #self.settings = QtCore.QSettings("Foundry", "ShotManager")
        self.settings = QtCore.QSettings(settings_path, QtCore.QSettings.IniFormat)
        self.settings.setFallbacksEnabled(False)
        self.buildUI()
        self.connectSignals()
        self.readSettings()
        self.label.setText('The settings are stored here:\n{}'.format(self.settings.fileName()))

    def readSettings(self):
        self.line_edit_widget.setText(self.settings.value('shotManager/xlsxLocation'))

    def writeSettings(self):
        self.settings.setValue('shotManager/xlsxLocation', self.line_edit_widget.text())

    def buildUI(self):
        self.load_btn = QtWidgets.QToolButton()
        self.load_btn.setText('...')
        self.line_edit_widget = QtWidgets.QLineEdit('')
        self.line_edit_widget.setReadOnly(True)
        self.label = QtWidgets.QLabel()
        self.label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        hLayout = QtWidgets.QHBoxLayout()
        vLayout = QtWidgets.QVBoxLayout()
        hLayout.addWidget(self.line_edit_widget)
        hLayout.addWidget(self.load_btn)
        vLayout.addLayout(hLayout)
        vLayout.addWidget(self.label)
        self.setLayout(vLayout)

    def connectSignals(self):
        self.load_btn.clicked.connect(self.openFileBrowser)
        self.line_edit_widget.editingFinished.connect(self.writeSettings)

    def openFileBrowser(self):
        file_path, filter = QtWidgets.QFileDialog.getOpenFileName(parent=self, caption='Open file', dir=self.line_edit_widget.text(), filter='Excel Spreadsheet (*.xlsx)')
        if file_path:
            self.line_edit_widget.setText(file_path)

class MenuWidget(QtWidgets.QMenu):
    """Custom menu to add actions to"""

    def __init__(self, title, parent=None):
        """Build the menu with custom actions"""
        super(MenuWidget, self).__init__(title, parent)

        # create actions
        write_shot_data_act = QtWidgets.QAction('Write Shot Data', self, triggered=self.write_shot_data_slot)
        read_shot_data_act = QtWidgets.QAction('Update Track Items', self, triggered=self.read_shot_data_slot)
        submit_for_review_act = QtWidgets.QAction('Submit for Client Review', self, triggered=self.submit_for_review)

        # add the actions to the menu
        self.addAction(write_shot_data_act)
        self.addAction(read_shot_data_act)
        self.addAction(submit_for_review_act)

    def read_shot_data_slot(self):
        """Open a filebrowser to read an xlsx file, then update track items with data found in the file's "shots" sheet"""
        db_handle = shot_manager_hiero.get_db_handle()
        shots_data = db_handle.read_shot_data('shots')
        shot_manager_hiero.update_track_items(shots_data)

    def submit_for_review(self):
        """Export track items as quicktime and send a thumbnail along with someversion info to the xlsx file.
        This assumes a preset with name 'Review Submission' to be present.
        """
        # get timeline selection
        selection = shot_manager_hiero.get_track_items()
        preset_name = 'Review Submission'
        preset = hiero.core.taskRegistry.processorPresetByName(preset_name)
        hiero.core.taskRegistry.createAndExecuteProcessor(preset, [hiero.core.ItemWrapper(i) for i in selection])        
    
    def write_shot_data_slot(self):
        """Open a filebrowser to point to a xlsx file, then write shot data to the file's "shots" sheet"""
        db_handle = shot_manager_hiero.get_db_handle()
        if not db_handle:
            return
        try:
            # write shot data to the spreadsheet and let the user know when it's done 
            db_handle.write_shot_data(shot_manager_hiero.collect_shot_data(), 'shots')
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText('Finished writing to:\n{}'.format(db_handle.xlsx_path))
        except:
            # inform the user of any error
            import sys
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText('An unexpected error ocurred:\n{}\n{}\n{}'.format(*sys.exc_info()))
        finally:
            # open message box
            msgBox.exec_()
    
    def add_to_menu(self, event):
        """Enable the menu to be registered as a new context menu item"""
        event.menu.addMenu(self)

